
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.012.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "A searchable store backed by Elasticsearch",
  "dist_author" => [
    "=over 4"
  ],
  "dist_name" => "Catmandu-Store-Elasticsearch",
  "dist_version" => "1.0",
  "license" => "perl",
  "module_name" => "Catmandu::Store::Elasticsearch",
  "recursive_test_files" => 1,
  "requires" => {
    "CQL::Parser" => "1.12",
    "Catmandu" => "1.04",
    "Cpanel::JSON::XS" => "3.0213",
    "Moo" => "1.004006",
    "Search::Elasticsearch" => "6.00",
    "namespace::clean" => "0.24",
    "perl" => "v5.10.1"
  },
  "test_requires" => {
    "Test::More" => "1.001003",
    "Test::Simple" => "1.001003"
  }
);


my %fallback_build_requires = (
  "Module::Build" => "0.28",
  "Test::More" => "1.001003",
  "Test::Simple" => "1.001003"
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
